/* $Id: wmfuzzy.c,v 1.26 2004/03/16 21:57:43 ian Exp $
 * Copyright (c) 2002-2003 Ian Glover
 *
 * This file is part of wmfuzzy and is distributed under the GNU
 * GPL. See file COPYING for details.
 */
#include "approximate.h"
#include "converttotext.h"
#include "display.h"
#include "logging.h"
#include "version.h"

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#define NAME "wmFuzzy"
#define MAX_LINE_LENGTH 9
#define MAX_NUM_LINES 4
/* Requires the array to be defined in the same scope as the call. */
#define ARRAY_SIZE(array) sizeof(array)/sizeof((array)[0])


Granularity gGranularity = FIVE;

static time_t get_time(void);
static void display_time(const struct tm* tm);
static void set_granularity(char* txt);

static void show_help(void);
static void show_version(void);

static void do_update(void);
static void do_click(void);

static void display_text(const char* text, size_t text_length);
static void down_case(char* txt);

time_t get_time(void) {
  struct timeval  tv;
  struct timezone tz;
  
  gettimeofday(&tv, &tz);
  return (tv.tv_sec);
}

void display_time(const struct tm* tm) {
  static char text[30];
  memset(text, 0, ARRAY_SIZE(text));  

  const ApproximateTime approximation = approximate(tm, gGranularity);

  if (approximation.minutes == 0)
  {
    if(gGranularity <= HOUR)
    {
      char buffer[30];
      memset(buffer, 0, ARRAY_SIZE(buffer));
      
      number_to_text(approximation.hour, buffer, ARRAY_SIZE(buffer));
      snprintf(text, ARRAY_SIZE(text), "%s o'clock", buffer);
    }
    else if(gGranularity == MONTHTHIRD)
    {
      char third_buffer[30];
      char month_buffer[30];
      memset(third_buffer, 0, ARRAY_SIZE(third_buffer));
      memset(month_buffer, 0, ARRAY_SIZE(month_buffer));

      
      monththird_to_text(approximation.monththird, third_buffer,
                         ARRAY_SIZE(third_buffer));
      month_to_text(approximation.month, month_buffer,
                    ARRAY_SIZE(month_buffer));
      snprintf(text, ARRAY_SIZE(text), "%s %s", third_buffer, month_buffer);
    }
    else
    {
      switch (gGranularity)
      {
        case SIXHOURS:
          hour_to_day_quarter(approximation.hour, text, ARRAY_SIZE(text));
          break;
        case TWELVEHOURS:
          hour_to_day_half(approximation.hour, text, ARRAY_SIZE(text));
          break;
        case DAY:
          weekday_to_text(approximation.weekday, text, ARRAY_SIZE(text));
          break;
        case WEEKTHIRD:
          weekthird_to_text(approximation.weekday, text, ARRAY_SIZE(text));
          break;          
        case MONTH:
          month_to_text(approximation.month, text, ARRAY_SIZE(text));
          break;
        case SEASON:
          season_to_text(approximation.month, text, ARRAY_SIZE(text));
          break;
        case YEAR:
          year_to_text(approximation.year, text, ARRAY_SIZE(text));
          break;
        case CENTURY:
          year_to_text(approximation.year, text, ARRAY_SIZE(text));
          break;
        default:
          assert(0);
          break;
      }
    }
  }
  else
  {
    char minutes[15];
    char hours[15];
    char* join;
    memset(minutes, 0, ARRAY_SIZE(minutes));
    memset(hours, 0, ARRAY_SIZE(hours));    
    
    if (approximation.join == TO) {
      join = "to";
    }
    else
    {
      assert(approximation.join == PAST);
      join = "past";
    }
    minute_to_text(approximation.minutes, minutes, ARRAY_SIZE(minutes));
    number_to_text(approximation.hour, hours, ARRAY_SIZE(hours));
    
    snprintf(text, ARRAY_SIZE(text), "%s %s %s", minutes, join, hours);
  }
  
  display_text(text, ARRAY_SIZE(text));
}

void display_text(const char* text, size_t text_length)
{
  int line_number = 0;
  size_t line_length;

  /* Text will be copied into raw_buffer and formatted correctly. +1
     is for the null character string terminator. */
  static char raw_buffer[(MAX_LINE_LENGTH + 1) * MAX_NUM_LINES];
  /* Lines will be built out of line_buffer. */
  static char line_buffer[(MAX_LINE_LENGTH + 1) * MAX_NUM_LINES];
  static char* line_text[MAX_NUM_LINES];
  
  const char* current = raw_buffer;
  const char* last = raw_buffer;
  int i = 0;

  memset(raw_buffer, 0, ARRAY_SIZE(raw_buffer));
  memset(line_buffer, 0, ARRAY_SIZE(line_buffer));
  memset(line_text, 0, ARRAY_SIZE(line_text));
  
  /* Set up the line pointrs into line_buffer. */
  for (i = 0; i < MAX_NUM_LINES; ++i )
  {
    line_text[i] = line_buffer + i * (MAX_LINE_LENGTH + 1);
  }
  
  /* text must be formatted to end with a space. Copy it into a new
     buffer to ensure this. */
  assert(strlen(text) <= text_length);
  assert(ARRAY_SIZE(raw_buffer) > text_length + 1);
  strcpy(raw_buffer, text);  
  assert(strlen(raw_buffer) < ARRAY_SIZE(raw_buffer) - 1);
  raw_buffer[strlen(raw_buffer)] = ' ';
  assert(strlen(raw_buffer) < ARRAY_SIZE(raw_buffer));  

  /* This has to be true to allow us space for the null characters to
     terminate the lines. */
  assert(ARRAY_SIZE(line_buffer) > text_length + MAX_NUM_LINES);
  /* Clear the line buffer. */  
  assert(line_buffer == line_text[0]);
  
  while (NULL != (current = strchr(last, ' ')))
  {
    assert(0 != current);
    assert(0 != last);
    /* assert(current > last); */
    if (!(current > last))
    {
      printf("%p %p %p ''%s''", current, last, raw_buffer, raw_buffer);
    }
    assert((size_t)(current - last) < text_length);
    assert(0 != line_text[line_number]);
           
    line_length = (size_t)(current - last);
    if (line_length > MAX_LINE_LENGTH)
    {
      line_length = MAX_LINE_LENGTH;
    }

    strncpy(line_text[line_number], last, line_length);

    assert(toupper(line_text[line_number][0]) < CHAR_MAX);

    line_text[line_number][0] = toupper(line_text[line_number][0]);

    last = current + 1;
    ++line_number;
  }
  set_message(line_text, line_number);
}

void down_case(char* txt)
{
  for ( ; *txt != '\0' ; ++txt)
  {
    *txt = tolower(*txt);
  }
}

void set_granularity(char* txt)
{
  long value;
  gGranularity = FIVE;

  if (txt == 0) { assert(0); return; }
  value = strtol(txt, 0, 10);

  down_case(txt);
  if (strcmp(txt, "century") == 0) { gGranularity = CENTURY; }
  else if (strcmp(txt, "year") == 0) { gGranularity = YEAR; }
  else if (strcmp(txt, "season") == 0) { gGranularity = SEASON; }
  else if (strcmp(txt, "month") == 0) { gGranularity = MONTH; }
  else if (strcmp(txt, "day") == 0) { gGranularity = DAY; }
  else
  {
    if (errno == ERANGE || errno == EINVAL)
    {
      warning("Error process -g argument value : %s\n", txt);
    }
    else if (value < 7) { gGranularity = FIVE; }
    else if (value < 12) { gGranularity = TEN; }
    else if (value < 17) { gGranularity = FIFTEEN; }
    else if (value < 22) { gGranularity = TWENTY; }
    else if (value < 45) { gGranularity = HALFHOUR; }
    else if (value < 210) { gGranularity = HOUR; }
    else if (value < 540) { gGranularity = SIXHOURS; }
    else if (value < 1080) { gGranularity = TWELVEHOURS; }
    else if (value < 2400) { gGranularity = DAY; }
    else if (value < 8880) { gGranularity = WEEKTHIRD; }
    else if (value < 28800) { gGranularity = MONTHTHIRD; }
    else if (value < 86400) { gGranularity = MONTH; }
    else if (value < 327600) { gGranularity = SEASON; }
    else if (value < 26542800) { gGranularity = YEAR; }
    else { gGranularity = CENTURY; }
  }
  
  /* fprintf(stderr, "Granularity set to %d\n", (int)gGranularity); */
}

void show_help(void) {
  show_version();
  
  printf(
    "\n"
    "-b<colour> :  Set the background colour. See -f.\n"
    "-f<colour> :  Set the foreground colour, either as a name or in the\n"
    "              form #RRGGBB where R,G and B are hexidecimal characters.\n"

    "-g<n>      :  Set the granularity, n should be one of 5, 10, 15, 20,\n"
    "              30, 60, 360, 720, 1440, 3360, 14400, 43200, 129600, 525600\n"
    "              or 52560000. Other values will be approximated to the\n"
    "              nearest of these. Alternatively the words \"century\",\n"
    "              \"year\", \"month\", or \"day\" can be used as shortcuts\n"
    "              for those settings. The default is 5 minutes.\n"
    "-h         :  Print this message and exit.\n"
    "-v         :  Print version information and exit.\n");
}

void show_version(void)
{
  printf("%s, version %s by Ian Glover (ian@manicai.net)\n", NAME, VERSION);
}

void do_update(void)
{
  time_t seconds;
  struct tm time;

  seconds = get_time();
  localtime_r(&seconds, &time);
  display_time(&time);
  set_update_delay(30);
}

void do_click(void)
{
  static char line_buffer[(MAX_NUM_LINES + 1) * MAX_LINE_LENGTH];
  static char* line_text[MAX_NUM_LINES];
  time_t seconds;
  struct tm time;
  int i = 0;

  memset(line_buffer, 0, ARRAY_SIZE(line_buffer));

  /* Get pointers into the line buffer. */
  for (i = 0; i < MAX_NUM_LINES; ++i)
  {
    line_text[i] = line_buffer + i * (MAX_LINE_LENGTH + 1);
  }

  seconds = get_time();
  localtime_r(&seconds, &time);

  assert(ARRAY_SIZE(line_text) >= 3);
  strftime(line_text[0], 9, "%H:%M", &time);
  strftime(line_text[1], 9, "%d %b", &time);
  strftime(line_text[2], 9, "%Y", &time);

  set_message(line_text, 3);
  set_update_delay(5);
}

void app_setup(int argc, char** argv) {
  int c;
  
  while( -1 != (c = getopt(argc, argv, "b:f:g:hv"))) {
    assert(c <= (int)CHAR_MAX);
    assert(c >= (int)CHAR_MIN);
      
    switch((char)c) {
      case '?': exit(EXIT_FAILURE);
      case 'b': set_background(optarg); break;
      case 'f': set_foreground(optarg); break;
      case 'h':
      {
        show_help();
        exit(EXIT_SUCCESS);
      }
      case 'g': set_granularity(optarg); break;
      case 'v':
      {
        show_version();
        exit(EXIT_SUCCESS);        
      }
      default:
      {
        fatal("Unhandled command line argument.");
        break;
      }
    }
  }

  set_update_callback(do_update);
  set_clicked_callback(do_click);

  set_app_name("wmFuzzy");
  
  set_update_delay(30); /* seconds */
}
